/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - data-table-columns.model.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {DataTableFieldTypeEnum} from "../../enums/datatable/data-table-field-type.enum";
import {OrderEnum} from "../../enums/order.enum";
import {DataTestPartialEnum} from "../../enums/partial/data-test-partial.enum";
import {ResourceNameSpace} from "../../stores/abstract/resource/resource-namespace.model";
import {ResourceState} from "../../stores/abstract/resource/resource.state";
import {ExtendEnum} from "../../types/extend-enum.type";
import {MappingObject} from "../../types/mapping-type.type";
import {BaseResourceWithLabels} from "../dto/base-resource-with-labels.model";
import {BaseResourceType} from "../dto/base-resource.model";
import {KeyValue} from "../key-value.model";
import {Sort} from "../query-parameters/sort.model";
import {DataTableActions} from "./data-table-actions.model";
import {DataTableComponent} from "./data-table-component.model";

export interface DataTableColumns<TResource extends BaseResourceType = any> {
  /**
   * The field to retrieve in row data
   */
  field: keyof TResource;
  /**
   * Label in header to translate
   */
  header: string;
  /**
   * Callback to display tooltip
   *
   * @deprecated not used
   * @param value
   */
  tooltip?: (value: string) => string | undefined;
  /**
   * Define how to handle the data, and what type of filter to use
   */
  type?: DataTableFieldTypeEnum;
  /**
   * The default sort to apply
   */
  order: OrderEnum;
  /**
   * List of enum to use to translate data. Also used in filter single select input. Only used if type = singleSelect
   */
  filterEnum?: KeyValue[] | BaseResourceWithLabels[];
  /**
   * Add data-test tag in the cell
   */
  dataTest?: ExtendEnum<DataTestPartialEnum>;
  /**
   * Translate cell data. Only used if type = singleSelect or string
   */
  translate?: boolean;
  /**
   * Component to inject. Use DataTableComponentHelper.get() static method
   */
  component?: DataTableComponent;
  /**
   * With of column
   */
  width?: string | undefined;
  /**
   * Min with of column
   */
  minWidth?: string | undefined;
  /**
   * Max with of column
   */
  maxWidth?: string | undefined;
  /**
   * Alignment of cell data
   */
  alignment?: "left" | "center" | "right";
  /**
   * Break word in cell data
   */
  breakWord?: boolean | undefined;
  /**
   * Apply text align center
   *
   * @deprecated use alignment
   */
  centerContent?: boolean;
  /**
   * Callback to add css class `is-highlighting-cell`
   */
  isHighlightCell?: ((data: TResource) => boolean) | undefined;

  /**
   * Define if the column is sortable
   */
  isSortable?: boolean;
  /**
   * Define if the column is filterable
   */
  isFilterable?: boolean;
  /**
   * Field used to filter
   * By default, filterable field used is `field`.
   * If you want to override, define here
   */
  filterableField?: keyof TResource;
  /**
   * Field used to sort
   * By default, sortable field used is `field`.
   * If you want to override, define here
   */
  sortableField?: keyof TResource;

  /**
   * Callback used to applay treatment in list item. Only used if type = list or listToJoinString
   */
  listCallback?: (value: any) => string;
  /**
   * Limit of item to display. Only used if type = list or listToJoinString
   */
  maxListItem?: number | undefined;
  /**
   * Callback to display conditionally cell
   */
  conditionallyDisplay?: ((data: TResource) => boolean) | undefined;
  /**
   * DataTableActions to display in cell. Only used if type = actionButton
   */
  action?: DataTableActions<TResource>;

  /**
   * Allow to use searchable single select. Only used if type = searchableSingleSelect
   */
  resourceNameSpace?: ResourceNameSpace;
  /**
   * Allow to use searchable single select. Only used if type = searchableSingleSelect
   */
  resourceState?: ResourceState<any, any>;
  /**
   * Allow to define sort criteria in searchable single select. Only used if type = searchableSingleSelect
   */
  searchableSingleSelectSort?: Sort;
  /**
   * Allow to define value in searchable single select. Only used if type = searchableSingleSelect
   */
  resourceValueKey?: string;
  /**
   * Allow to define label to use in searchable single select. Only used if type = searchableSingleSelect
   */
  resourceLabelKey?: string;
  /**
   * Allow to define if `resourceLabelCallback` should be used in searchable single select. Only used if type = searchableSingleSelect
   *
   * @deprecated not used
   */
  useResourceLabelCallback?: boolean | undefined;
  /**
   * Allow to use callback to generate label in searchable single select. Only used if type = searchableSingleSelect
   */
  resourceLabelCallback?: (value: any) => string;
  /**
   * Allow to use a callback to display a second line in searchable single select. Only used if type = searchableSingleSelect
   */
  resourceExtraLabelSecondLineCallback?: (value: any) => string;
  /**
   * Allow to define extra query param in request throw by searchable single select. Only used if type = searchableSingleSelect
   */
  extraSearchQueryParam?: MappingObject<string, string> | undefined;

  /**
   * @deprecated not used
   */
  isUser?: boolean;
}
