/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - composition.state.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {Type} from "@angular/core";
import {
  Actions,
  StateContext,
  Store,
} from "@ngxs/store";
import {Observable} from "rxjs";
import {
  catchError,
  tap,
} from "rxjs/operators";
import {SOLIDIFY_CONSTANTS} from "../../../constants";
import {RegisterDefaultAction} from "../../../decorators/store.decorator";
import {ModelAttributeEnum} from "../../../enums/model-attribute.enum";
import {ApiResourceNamePartialEnum} from "../../../enums/partial/api-resource-name-partial.enum";
import {SolidifyStateError} from "../../../errors/solidify-state.error";
import {ApiService} from "../../../http/api.service";
import {BaseResourceType} from "../../../models/dto/base-resource.model";
import {CollectionTyped} from "../../../models/dto/collection-typed.model";
import {SolidifyHttpErrorResponseModel} from "../../../models/errors/solidify-http-error-response.model";
import {QueryParameters} from "../../../models/query-parameters/query-parameters.model";
import {NotifierService} from "../../../models/services/notifier-service.model";
import {BaseStateModel} from "../../../models/stores/base-state.model";
import {
  isFalse,
  isNullOrUndefined,
  isTrue,
} from "../../../tools/is/is.tool";
import {ExtendEnum} from "../../../types/extend-enum.type";
import {ObjectUtil} from "../../../utils/object.util";
import {MemoizedUtil} from "../../../utils/stores/memoized.util";
import {StoreUtil} from "../../../utils/stores/store.util";
import {BaseOptions} from "../base/base-options.model";
import {
  BaseState,
  defaultBaseStateInitValue,
} from "../base/base.state";
import {CompositionActionHelper} from "./composition-action.helper";
import {CompositionNameSpace} from "./composition-namespace.model";
import {CompositionOptions} from "./composition-options.model";
import {CompositionStateModel} from "./composition-state.model";
import {CompositionAction} from "./composition.action";

export const defaultCompositionStateInitValue: <TResourceType extends BaseResourceType> () => CompositionStateModel<TResourceType> = () =>
  ({
    ...defaultBaseStateInitValue(),
    total: 0,
    list: undefined,
    current: undefined,
    queryParameters: new QueryParameters(),
  });

// @dynamic
export abstract class CompositionState<TStateModel extends BaseStateModel, TResource extends BaseResourceType> extends BaseState<TStateModel> {
  protected readonly _resourceName: ExtendEnum<ApiResourceNamePartialEnum>;
  protected override readonly _nameSpace: CompositionNameSpace;
  protected override readonly _optionsState: CompositionOptions;

  protected constructor(protected _apiService: ApiService,
                        protected _store: Store,
                        protected _notificationService: NotifierService,
                        protected _actions$: Actions,
                        protected _options: CompositionOptions) {
    super(_apiService, _store, _notificationService, _actions$, _options, CompositionState);
    this._resourceName = this._optionsState.resourceName;
  }

  protected static _getDefaultOptions(): CompositionOptions | any {
    let defaultOptions: CompositionOptions | any = {
      keepCurrentStateAfterCreate: false,
      keepCurrentStateAfterUpdate: false,
      keepCurrentStateAfterDelete: false,
    } as CompositionOptions;
    defaultOptions = Object.assign(BaseState._getDefaultOptions(), defaultOptions);
    return defaultOptions;
  }

  protected static override _checkOptions(stateName: string, options: BaseOptions): void {
    BaseState._checkOptions(stateName, options);
  }

  static list<TStateModel extends CompositionStateModel<TResource>, TResource extends BaseResourceType>(store: Store, ctor: Type<CompositionState<TStateModel, TResource>>): Observable<TResource[]> {
    return MemoizedUtil.select(store, ctor, state => state.list, true);
  }

  static listSnapshot<TStateModel extends CompositionStateModel<TResource>, TResource extends BaseResourceType>(store: Store, ctor: Type<CompositionState<TStateModel, TResource>>): TResource[] {
    return MemoizedUtil.selectSnapshot(store, ctor, state => state.list);
  }

  static current<TStateModel extends CompositionStateModel<TResource>, TResource extends BaseResourceType>(store: Store, ctor: Type<CompositionState<TStateModel, TResource>>): Observable<TResource> {
    return MemoizedUtil.select(store, ctor, state => state.current, true);
  }

  static currentSnapshot<TStateModel extends CompositionStateModel<TResource>, TResource extends BaseResourceType>(store: Store, ctor: Type<CompositionState<TStateModel, TResource>>): TResource {
    return MemoizedUtil.selectSnapshot(store, ctor, state => state.current);
  }

  static total<TStateModel extends CompositionStateModel<TResource>, TResource extends BaseResourceType>(store: Store, ctor: Type<CompositionState<TStateModel, TResource>>): Observable<number> {
    return MemoizedUtil.select(store, ctor, state => state.total, true);
  }

  static totalSnapshot<TStateModel extends CompositionStateModel<TResource>, TResource extends BaseResourceType>(store: Store, ctor: Type<CompositionState<TStateModel, TResource>>): number {
    return MemoizedUtil.selectSnapshot(store, ctor, state => state.total);
  }

  static queryParameters<TStateModel extends CompositionStateModel<TResource>, TResource extends BaseResourceType>(store: Store, ctor: Type<CompositionState<TStateModel, TResource>>): Observable<QueryParameters> {
    return MemoizedUtil.select(store, ctor, state => state.queryParameters, true);
  }

  static queryParametersSnapshot<TStateModel extends CompositionStateModel<TResource>, TResource extends BaseResourceType>(store: Store, ctor: Type<CompositionState<TStateModel, TResource>>): QueryParameters {
    return MemoizedUtil.selectSnapshot(store, ctor, state => state.queryParameters);
  }

  protected _evaluateSubResourceUrl(parentId: string): string {
    return this._urlResource + SOLIDIFY_CONSTANTS.URL_SEPARATOR + parentId + SOLIDIFY_CONSTANTS.URL_SEPARATOR + this._resourceName;
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.ChangeQueryParameters)
  changeQueryParameters(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.ChangeQueryParameters): void {
    ctx.patchState({
      queryParameters: action.queryParameters,
    });
    if (isTrue(action.getAllAfterChange)) {
      ctx.dispatch(CompositionActionHelper.getAll(this._nameSpace, action.parentId, undefined, action.keepCurrentContext));
    }
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.GetAll)
  getAll<U>(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.GetAll): Observable<CollectionTyped<U>> {
    let reset = {};
    if (!action.keepCurrentContext) {
      reset = {
        total: 0,
        list: undefined,
      };
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      queryParameters: StoreUtil.getQueryParametersToApply(action.queryParameters, ctx),
      ...reset,
    });
    const url = this._evaluateSubResourceUrl(action.parentId);
    return this._apiService.getCollection<U>(url, ctx.getState().queryParameters)
      .pipe(
        tap((collection: CollectionTyped<U>) => {
          ctx.dispatch(CompositionActionHelper.getAllSuccess<U>(this._nameSpace, action, collection));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(CompositionActionHelper.getAllFail(this._nameSpace, action));
          throw new SolidifyStateError(this, error);
        }),
      );
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.GetAllSuccess)
  getAllSuccess(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.GetAllSuccess<TResource>): void {
    const queryParameters = StoreUtil.updateQueryParameters(ctx, action.list);

    ctx.patchState({
      total: action.list._page?.totalItems,
      list: action.list._data,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      queryParameters: queryParameters,
    });
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.GetAllFail)
  getAllFail(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.GetAllFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.GetById)
  getById<U>(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.GetById): Observable<U> {
    let reset = {};
    if (!action.keepCurrentContext) {
      reset = {
        current: undefined,
      };
    }
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
      ...reset,
    });
    const url = this._evaluateSubResourceUrl(action.parentId);
    return this._apiService.getById<U>(url, action.resId)
      .pipe(
        tap((model: U) => {
          ctx.dispatch(CompositionActionHelper.getByIdSuccess<U>(this._nameSpace, action, model));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(CompositionActionHelper.getByIdFail(this._nameSpace, action));
          throw new SolidifyStateError(this, error);
        }),
      );
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.GetByIdSuccess)
  getByIdSuccess(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.GetByIdSuccess<TResource>): void {
    ctx.patchState({
      current: action.model,
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.GetByIdFail)
  getByIdFail(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.GetByIdFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.Create)
  create(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.Create<TResource>): Observable<TResource> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });
    const url = this._evaluateSubResourceUrl(action.parentId);
    return this._apiService.post<TResource>(url, action.model)
      .pipe(
        tap((model: TResource) => {
          ctx.dispatch(CompositionActionHelper.createSuccess(this._nameSpace, action, action.parentId, model));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(CompositionActionHelper.createFail(this._nameSpace, action));
          throw new SolidifyStateError(this, error);
        }),
      );
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.CreateSuccess)
  createSuccess(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.CreateSuccess<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      current: action.model,
    });
    StoreUtil.navigateCompositionIfDefined(ctx, this._optionsState.routeRedirectUrlAfterSuccessCreateAction, action.parentId, action.model.resId, this._optionsState.routeReplaceAfterSuccessCreateAction);
    StoreUtil.notifySuccess(this._notificationService, this._optionsState.notificationResourceCreateSuccessTextToTranslate);
    this._cleanCurrentStateIfDefined(ctx, this._optionsState.keepCurrentStateAfterCreate);
    ctx.dispatch(CompositionActionHelper.getAll(this._nameSpace, action.parentId));
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.CreateFail)
  createFail(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.CreateFail<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
    StoreUtil.notifyError(this._notificationService, this._optionsState.notificationResourceCreateFailTextToTranslate);
    this._cleanCurrentStateIfDefined(ctx, this._optionsState.keepCurrentStateAfterCreate);
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.Delete)
  delete(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.Delete): Observable<TResource> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });
    const url = this._evaluateSubResourceUrl(action.parentId);
    return this._apiService.deleteById<TResource>(url, action.resId)
      .pipe(
        tap(() => {
          ctx.dispatch(CompositionActionHelper.deleteSuccess(this._nameSpace, action, action.parentId));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(CompositionActionHelper.deleteFail(this._nameSpace, action));
          throw new SolidifyStateError(this, error);
        }),
      );
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.DeleteSuccess)
  deleteSuccess(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.DeleteSuccess): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
    StoreUtil.navigateCompositionIfDefined(ctx, this._optionsState.routeRedirectUrlAfterSuccessDeleteAction, action.parentId, undefined, this._optionsState.routeReplaceAfterSuccessDeleteAction);
    StoreUtil.notifySuccess(this._notificationService, this._optionsState.notificationResourceDeleteSuccessTextToTranslate);
    this._cleanCurrentStateIfDefined(ctx, this._optionsState.keepCurrentStateAfterDelete);
    ctx.dispatch(CompositionActionHelper.getAll(this._nameSpace, action.parentId));
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.DeleteFail)
  deleteFail(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.DeleteFail): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
    StoreUtil.notifyError(this._notificationService, this._optionsState.notificationResourceDeleteFailTextToTranslate);
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.Update)
  update(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.Update<TResource>): Observable<TResource> {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter + 1,
    });
    const url = this._evaluateSubResourceUrl(action.parentId);
    return this._apiService.patchById<TResource>(url, action.model[ModelAttributeEnum.resId], action.model)
      .pipe(
        tap((model: TResource) => {
          ctx.dispatch(CompositionActionHelper.updateSuccess(this._nameSpace, action, action.parentId, model));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(CompositionActionHelper.updateFail(this._nameSpace, action, action.parentId));
          throw new SolidifyStateError(this, error);
        }),
      );
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.UpdateSuccess)
  updateSuccess(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.UpdateSuccess<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
      current: action.model,
    });
    StoreUtil.navigateCompositionIfDefined(ctx, this._optionsState.routeRedirectUrlAfterSuccessUpdateAction, action.parentId, action.model.resId, this._optionsState.routeReplaceAfterSuccessUpdateAction);
    StoreUtil.notifySuccess(this._notificationService, this._optionsState.notificationResourceUpdateSuccessTextToTranslate);
    this._cleanCurrentStateIfDefined(ctx, this._optionsState.keepCurrentStateAfterUpdate);
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.UpdateFail)
  updateFail(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.UpdateFail<TResource>): void {
    ctx.patchState({
      isLoadingCounter: ctx.getState().isLoadingCounter - 1,
    });
    StoreUtil.notifyError(this._notificationService, this._optionsState.notificationResourceUpdateFailTextToTranslate);
    ctx.dispatch(CompositionActionHelper.getAll(this._nameSpace, action.parentId));
  }

  @RegisterDefaultAction((compositionNameSpace: CompositionNameSpace) => compositionNameSpace.Clean)
  clean(ctx: StateContext<CompositionStateModel<TResource>>, action: CompositionAction.Clean): void {
    this._resetStateToDefault(ctx, isNullOrUndefined(action.preserveList) ? true : action.preserveList, action.queryParameter);
  }

  private _cleanCurrentStateIfDefined(ctx: StateContext<CompositionStateModel<TResource>>, keepCurrentState: boolean): void {
    if (isNullOrUndefined(keepCurrentState) || isFalse(keepCurrentState)) {
      ctx.dispatch(CompositionActionHelper.clean(this._nameSpace, true));
    }
  }

  private _resetStateToDefault(ctx: StateContext<CompositionStateModel<TResource>>, preserveListData: boolean, newQueryParameters?: QueryParameters): CompositionStateModel<TResource> {
    const oldList = ctx.getState().list;
    const oldTotal = ctx.getState().total;
    const oldQueryParameters = ctx.getState().queryParameters;
    const newState = this._getDefaultData();
    if (isTrue(preserveListData)) {
      newState.list = oldList;
      newState.total = oldTotal;
      newState.queryParameters = oldQueryParameters;
    }
    if (!isNullOrUndefined(newQueryParameters)) {
      newState.queryParameters = newQueryParameters;
    }
    return ctx.patchState(newState);
  }

  private _getDefaultData(): CompositionStateModel<TResource> {
    return ObjectUtil.clone(StoreUtil.getDefaultData(this.constructor), true);
  }
}
