/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - resource-logo.state.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  HttpClient,
  HttpEventType,
  HttpHeaders,
} from "@angular/common/http";
import {Inject} from "@angular/core";
import {
  Actions,
  StateContext,
  Store,
} from "@ngxs/store";
import {Observable} from "rxjs";
import {
  catchError,
  map,
  tap,
} from "rxjs/operators";
import {RegisterDefaultAction} from "../../../decorators/store.decorator";
import {ApiActionNamePartialEnum} from "../../../enums/partial/api-action-name-partial.enum";
import {DefaultSolidifyEnvironment} from "../../../environments/environment.solidify-defaults";
import {ApiService} from "../../../http/api.service";
import {ENVIRONMENT} from "../../../injection-tokens/environment.injection-token";
import {BaseResourceAvatar} from "../../../models/dto/base-resource-avatar.model";
import {BaseResourceLogo} from "../../../models/dto/base-resource-logo.model";
import {BaseResourceThumbnail} from "../../../models/dto/base-resource-thumbnail.model";
import {BaseResourceWithLogo} from "../../../models/dto/base-resource-with-logo.model";
import {Logo} from "../../../models/dto/logo.model";
import {SolidifyHttpErrorResponseModel} from "../../../models/errors/solidify-http-error-response.model";
import {UploadEventModel} from "../../../models/http/upload-event.model";
import {BaseStateModel} from "../../../models/stores/base-state.model";
import {ActionSubActionCompletionsWrapper} from "../../../models/stores/base.action";
import {NotificationService} from "../../../services/notification.service";
import {
  isNotNullNorUndefined,
  isNullOrUndefined,
} from "../../../tools/is/is.tool";
import {MappingObjectUtil} from "../../../utils/mapping-object.util";
import {ObjectUtil} from "../../../utils/object.util";
import {ofSolidifyActionCompleted} from "../../../utils/stores/store.tool";
import {StringUtil} from "../../../utils/string.util";
import {ResourceAction} from "../resource/resource.action";
import {
  defaultResourceStateInitValue,
  ResourceState,
} from "../resource/resource.state";
import {ResourceLogoActionHelper} from "./resource-logo-action.helper";
import {ResourceLogoNameSpace} from "./resource-logo-namespace.model";
import {ResourceLogoOptions} from "./resource-logo-options.model";
import {ResourceLogoStateModel} from "./resource-logo-state.model";
import {ResourceLogoAction} from "./resource-logo.action";

export const defaultResourceLogoStateInitValue: <TResourceType extends BaseResourceWithLogo> () => ResourceLogoStateModel<TResourceType> = () =>
  ({
    ...defaultResourceStateInitValue(),
    logo: undefined,
    isLoadingLogo: false,
    listLogo: {},
    isLoadingAllLogo: false,
  });

export abstract class ResourceLogoState<TStateModel extends BaseStateModel, TResource extends BaseResourceWithLogo> extends ResourceState<TStateModel, TResource> {
  protected readonly _FILE_KEY: string = "file";

  protected readonly _nameSpace: ResourceLogoNameSpace;

  protected constructor(protected _apiService: ApiService,
                        protected _store: Store,
                        protected _notificationService: NotificationService,
                        protected _actions$: Actions,
                        protected _options: ResourceLogoOptions,
                        protected _httpClient: HttpClient,
                        protected _mode: ResourceLogoStateModeEnum,
                        @Inject(ENVIRONMENT) private readonly _environment: DefaultSolidifyEnvironment) {
    super(_apiService, _store, _notificationService, _actions$, _options);
  }

  protected _getDownloadAction(): ApiActionNamePartialEnum {
    if (this._mode === ResourceLogoStateModeEnum.logo) {
      return this._environment.resourceLogoApiActionNameDownloadLogo;
    }
    if (this._mode === ResourceLogoStateModeEnum.avatar) {
      return this._environment.resourceLogoApiActionNameDownloadAvatar;
    }
    if (this._mode === ResourceLogoStateModeEnum.thumbnail) {
      return this._environment.resourceLogoApiActionNameDownloadThumbnail;
    }
  }

  protected _getUploadAction(): ApiActionNamePartialEnum {
    if (this._mode === ResourceLogoStateModeEnum.logo) {
      return this._environment.resourceLogoApiActionNameUploadLogo;
    }
    if (this._mode === ResourceLogoStateModeEnum.avatar) {
      return this._environment.resourceLogoApiActionNameUploadAvatar;
    }
    if (this._mode === ResourceLogoStateModeEnum.thumbnail) {
      return this._environment.resourceLogoApiActionNameUploadThumbnail;
    }
  }

  protected _getDeleteAction(): ApiActionNamePartialEnum {
    if (this._mode === ResourceLogoStateModeEnum.logo) {
      return this._environment.resourceLogoApiActionNameDeleteLogo;
    }
    if (this._mode === ResourceLogoStateModeEnum.avatar) {
      return this._environment.resourceLogoApiActionNameDeleteAvatar;
    }
    if (this._mode === ResourceLogoStateModeEnum.thumbnail) {
      return this._environment.resourceLogoApiActionNameDeleteThumbnail;
    }
  }

  protected _getLogo(model: TResource): Logo {
    if (this._mode === ResourceLogoStateModeEnum.logo) {
      return (model as BaseResourceLogo)?.logo;
    }
    if (this._mode === ResourceLogoStateModeEnum.avatar) {
      return (model as BaseResourceAvatar)?.avatar;
    }
    if (this._mode === ResourceLogoStateModeEnum.thumbnail) {
      return (model as BaseResourceThumbnail)?.thumbnail;
    }
  }

  protected abstract get _urlLogoResource(): string;

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.GetPhoto, {}, ResourceState)
  getPhoto(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.GetPhoto): Observable<any> {
    this._isLoadingLogo(ctx, true);

    let headers = new HttpHeaders();
    headers = headers.set("Content-Disposition", "attachment; filename=" + "photo");

    return this._httpClient.get(`${this._urlLogoResource}/${action.resId}/${this._getDownloadAction()}`, {
      headers,
      responseType: "blob",
    }).pipe(
      tap((data: Blob) => {
        ctx.dispatch(ResourceLogoActionHelper.getPhotoSuccess(this._nameSpace, action, data));
      }),
      catchError((error: SolidifyHttpErrorResponseModel) => {
        ctx.dispatch(ResourceLogoActionHelper.getPhotoFail(this._nameSpace, action));
        throw error;
      }),
    );
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.GetPhotoSuccess, {}, ResourceState)
  getPhotoSuccess(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.GetPhotoSuccess): void {
    ctx.patchState({
      isLoadingLogo: false,
      logo: isNullOrUndefined(action.blob) ? undefined : URL.createObjectURL(action.blob),
    });
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.GetPhotoFail, {}, ResourceState)
  getPhotoFail(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.GetPhotoFail): void {
    ctx.patchState({
      isLoadingLogo: false,
      logo: undefined,
    });
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.GetPhotoByResId, {}, ResourceState)
  getPhotoByResId(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.GetPhotoByResId): Observable<any> {
    ctx.patchState({
      isLoadingAllLogo: true,
    });

    let headers = new HttpHeaders();
    headers = headers.set("Content-Disposition", "attachment; filename=" + "photo");

    return this._httpClient.get(`${this._urlLogoResource}/${action.resId}/${this._getDownloadAction()}`, {
      headers,
      responseType: "blob",
    }).pipe(
      tap((data: Blob) => {
        ctx.dispatch(ResourceLogoActionHelper.getPhotoByResIdSuccess(this._nameSpace, action, data));
      }),
      catchError((error: SolidifyHttpErrorResponseModel) => {
        ctx.dispatch(ResourceLogoActionHelper.getPhotoByResIdFail(this._nameSpace, action));
        throw error;
      }),
    );
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.GetPhotoByResIdSuccess, {}, ResourceState)
  getPhotoByResIdSuccess(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.GetPhotoByResIdSuccess): void {
    const logoCopy = ObjectUtil.clone(ctx.getState().listLogo);
    if (isNullOrUndefined(action.blob)) {
      MappingObjectUtil.set(logoCopy, action.parentAction.resId, StringUtil.stringEmpty);
    } else {
      MappingObjectUtil.set(logoCopy, action.parentAction.resId, URL.createObjectURL(action.blob));
    }
    ctx.patchState({
      isLoadingAllLogo: false,
      listLogo: logoCopy,
    });
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.GetPhotoByResIdFail, {}, ResourceState)
  getPhotoByResIdFail(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.GetPhotoByResIdFail): void {
    const logoCopy = ObjectUtil.clone(ctx.getState().listLogo);
    MappingObjectUtil.set(logoCopy, action.parentAction.resId, StringUtil.stringEmpty);

    ctx.patchState({
      isLoadingAllLogo: false,
      listLogo: logoCopy,
    });
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.UploadPhoto, {}, ResourceState)
  uploadPhoto(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.UploadPhoto): Observable<any> {
    this._isLoadingLogo(ctx, true);
    const formData = new FormData();
    formData.append(this._FILE_KEY, action.file, action.file.name);

    return this._apiService.upload(`${this._urlLogoResource}/${action.resId}/${this._getUploadAction()}`, formData)
      .pipe(
        map((event: UploadEventModel) => {
          switch (event.type) {
            case HttpEventType.DownloadProgress:
            case HttpEventType.UploadProgress:
              return;
            case HttpEventType.Response:
              if (isNotNullNorUndefined(event.body) &&
                (
                  (this._mode === ResourceLogoStateModeEnum.avatar && isNotNullNorUndefined((event.body as BaseResourceAvatar).avatar)) ||
                  (this._mode === ResourceLogoStateModeEnum.logo && isNotNullNorUndefined((event.body as BaseResourceLogo).logo)) ||
                  (this._mode === ResourceLogoStateModeEnum.thumbnail && isNotNullNorUndefined((event.body as BaseResourceThumbnail).thumbnail))
                )
              ) {
                ctx.dispatch(ResourceLogoActionHelper.uploadPhotoSuccess(this._nameSpace, action));
              } else {
                ctx.dispatch(ResourceLogoActionHelper.getPhotoFail(this._nameSpace, action));
              }
              return;
            default:
              return;
          }
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(ResourceLogoActionHelper.uploadPhotoFail(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.UploadPhotoSuccess, {}, ResourceState)
  uploadPhotoSuccess(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.UploadPhotoSuccess): void {
    this._isLoadingLogo(ctx, false);
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.UploadPhotoFail, {}, ResourceState)
  uploadPhotoFail(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.UploadPhotoFail): void {
    this._isLoadingLogo(ctx, false);
  }

  protected _isLoadingLogo(ctx: StateContext<ResourceLogoStateModel<TResource>>, isLoading: boolean): void {
    ctx.patchState({
      isLoadingLogo: isLoading,
    });
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.DeletePhoto, {}, ResourceState)
  deletePhoto(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.DeletePhoto): Observable<any> {
    this._isLoadingLogo(ctx, true);

    return this._apiService.delete(`${this._urlLogoResource}/${action.resId}/${this._getDeleteAction()}`)
      .pipe(
        tap((result: any) => {
          ctx.dispatch(ResourceLogoActionHelper.deletePhotoSuccess(this._nameSpace, action));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(ResourceLogoActionHelper.deletePhotoSuccess(this._nameSpace, action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.DeletePhotoSuccess, {}, ResourceState)
  deletePhotoSuccess(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.DeletePhotoSuccess): void {
    ctx.patchState({
      isLoadingLogo: false,
      logo: undefined,
    });
  }

  @RegisterDefaultAction((userAvatarNameSpace: ResourceLogoNameSpace) => userAvatarNameSpace.DeletePhotoFail, {}, ResourceState)
  deletePhotoFail(ctx: StateContext<ResourceLogoStateModel<TResource>>, action: ResourceLogoAction.DeletePhotoFail): void {
    this._isLoadingLogo(ctx, false);
  }

  protected override _getListActionsUpdateSubResource(model: TResource, action: ResourceAction.Create<TResource> | ResourceAction.Update<TResource>, ctx: StateContext<ResourceLogoStateModel<TResource>>): ActionSubActionCompletionsWrapper[] {
    const actions = super._getListActionsUpdateSubResource(model, action, ctx);
    actions.push(...this._getActionUpdateLogo(model, action, ctx));
    return actions;
  }

  protected _getActionUpdateLogo(model: TResource, action: ResourceAction.Create<TResource> | ResourceAction.Update<TResource>, ctx: StateContext<ResourceLogoStateModel<TResource>>): ActionSubActionCompletionsWrapper[] {
    return ResourceLogoState.getActionUpdateLogo(model.resId, action.modelFormControlEvent.model.newPendingLogo, this._getLogo(model), this._nameSpace, this._actions$);
  }

  static getActionUpdateLogo(resourceId: string, newPendingLogo: Blob | null | undefined, oldLogo: Logo | null | undefined, nameSpace: ResourceLogoNameSpace, _actions$: Actions): ActionSubActionCompletionsWrapper[] {
    const actions = [];

    const isLogoToUpload = isNotNullNorUndefined(newPendingLogo);
    const isLogoToDelete = newPendingLogo === null;
    if (isLogoToUpload) {
      actions.push({
        action: ResourceLogoActionHelper.uploadPhoto(nameSpace, resourceId, newPendingLogo as any),
        subActionCompletions: [
          _actions$.pipe(ofSolidifyActionCompleted(nameSpace.UploadPhotoSuccess)),
          _actions$.pipe(ofSolidifyActionCompleted(nameSpace.UploadPhotoFail)),
        ],
      });
    } else if (isLogoToDelete) {
      if (isNotNullNorUndefined(oldLogo)) {
        actions.push({
          action: ResourceLogoActionHelper.deletePhoto(nameSpace, resourceId, null),
          subActionCompletions: [
            _actions$.pipe(ofSolidifyActionCompleted(nameSpace.DeletePhotoSuccess)),
            _actions$.pipe(ofSolidifyActionCompleted(nameSpace.DeletePhotoFail)),
          ],
        });
      }
    }
    return actions;
  }
}

export enum ResourceLogoStateModeEnum {
  avatar = "avatar",
  logo = "logo",
  thumbnail = "thumbnail",
  dataset = "dataset",
}
